/* --- Copyright University of Sussex 1991.  All rights reserved. ---------
 > File:          C.all/lib/auto/fi_check.p
 > Purpose:       FAST Check that an object is an integer in a range.
 > Author:        Aaron Sloman May 1990 (see revisions)
 > Documentation: HELP * CHECKINTEGER
 > Related Files: LIB * CHECKINTEGER
 */
compile_mode :pop11 +strict :vm -pentch;

section;

define fi_check(obj, low, high) with_props false;
	;;; returns the obj if it is an integer and >=low and <= high.
	;;; Otherwise causes an error
	lvars obj low high;
	;;; This version does not check that low and high are integers
	if obj.isinteger then
		if low then
			if obj fi_< low then ;;; error
			elseif high then
				if obj fi_> high then ;;; error
				else
					return(obj)	;;; within bounds
				endif
			else
				return(obj);	;;; no upper limit
			endif
		elseif high then
			if obj fi_> high then ;;; error
			else
				return(obj) ;;; no lower bound, and within upper
			endif
		else
			return(obj)	;;; no limits specified
		endif;
	else
		;;; error
	endif;
	mishap(obj, 1,
		if low and high then
			'INTEGER BETWEEN ' sys_>< low sys_>< ' AND ' sys_>< high sys_>< ' NEEDED'
		elseif low then
			'INTEGER GREATER OR EQUAL TO ' sys_>< low sys_>< ' NEEDED'
		elseif high then
			'INTEGER LESS OR EQUAL TO ' sys_>< high sys_>< ' NEEDED'
		else
			'INTEGER NEEDED'
		endif);
enddefine;

endsection;


/* --- Revision History ---------------------------------------------------
--- John Gibson, Oct 12 1991
		Changed mishap message
--- Aaron Sloman, May 23 1990
		Optimised version of checkinteger (because compiler is not too good
		on complex boolean expressions). Also returns the integer
 */
