/*--- Copyright University of Sussex 1995. All rights reserved. ----------
 > File:            C.all/lib/auto/postscript_line_consumer.p
 > Purpose:         PostScript (d)string line consumer
 > Author:          Jonathan Meyer, Apr 1 1993 (see revisions)
 > Documentation:   REF * POSTSCRIPT_LINE_CONSUMER
 > Related Files:
 */
section;
compile_mode :pop11 +strict;

/*
postscript_line_consumer(DEVICE) -> CONS_P
		Creates a consumer which takes (d)strings and generates postscript
		code for them.
*/

include vedscreendefs;

;;; The following PostScript header is printed at the start of each document.
;;; It defines DSC comments that structure the document to conform with
;;; DSC 3.0. It also defines the PostScript code used to setup the fonts.

;;; (we use a vector of strings so we can insert comments into it.
;;; however, to save space, we coerce this into one long string.)

lconstant preamble = nonwriteable consstring(#|appdata({

'%%DocumentNeededResources: font Courier Courier-Bold'
'%%+ font Courier-Oblique Courier-BoldOblique'

'%%DocumentSuppliedResources: font Courier-ISOLatin1 Courier-ISOLatin1-Bold'
'%%+ font Courier-ISOLatin1-Oblique Courier-ISOLatin1-BoldOblique'
'%%+ font VT100Graphics VT100Graphics-Bold'

'%%Orientation: Portrait'
'%%PageOrder: Ascend'
'%%Pages: (atend)'
'%%EndComments'

'%%BeginProlog'

'/docsave save def'

;;; declare various shorthand macros
'/Bd {bind def} bind def'
'/Mt {moveto} Bd'
'/Lt {lineto} Bd'
'/Sg {setgray} Bd'
'/Slw {setlinewidth} Bd'
'/Sh {show} Bd'

;;; define SF function to use selectfont if it is available
'/languagelevel where'
'{pop languagelevel} {1} ifelse'
'2 lt {'
'/SF { exch findfont exch dup type /arraytype eq {makefont} {scalefont} ifelse'
'   setfont } Bd'
'} {'
'/SF /selectfont load def'
'} ifelse'

;;; function used to draw characters of VT100 graphics font

'/HVLine {'
'   /vl exch def /hl exch def'
'   newpath'
'   hl 0 ne {'
'       hl 2 eq {300} {0} ifelse 300 Mt'
'       hl 1 eq {300} {600} ifelse 300 Lt'
'   } if'
'   vl 0 ne {'
'       300 vl 2 eq {300} {800} ifelse Mt'
'       300 vl 1 eq {300} {-200} ifelse Lt'
'   } if'
'   stroke'
'} Bd'

;;; Creates a font for drawing VED vt100 graphics characters
;;; Usage: <fontname> <linewidth> DefVT100GraphicsFont

'/DefVT100GraphicsFont {'
'   10 dict begin'
'       /FontType 3 def'
'       /FontMatrix [.001 0 0 .001 0 0] def'
'       /FontBBox [0 0 1000 1000] def'
'       /Encoding 256 array def'
'       /LineWidth exch def'
'       0 1 255 {Encoding exch /.notdef put} for'
'       Encoding'
'       dup 97  /leftend put'
'       dup 98  /rightend put'
'       dup 99  /hline put'
'       dup 100 /bottomend put'
'       dup 101 /bottomleft put'
'       dup 102 /bottomright put'
'       dup 103 /bottomtee put'
'       dup 104 /topend put'
'       dup 105 /topleft put'
'       dup 106 /topright put'
'       dup 107 /toptee put'
'       dup 108 /vline put'
'       dup 109 /lefttee put'
'       dup 110 /righttee put'
'       dup 111 /cross put'
'       dup 113 /diamondsign put'
'       pop'
'       /CharProcs 19 dict def'
'       CharProcs begin'
'           /.notdef {} def'
'           /leftend     { 3 0 HVLine } Bd'
'           /rightend    { 3 0 HVLine } Bd'
'           /topend      { 0 3 HVLine } Bd'
'           /bottomend   { 0 3 HVLine } Bd'
'           /topleft     { 2 2 HVLine } Bd'
'           /topright    { 1 2 HVLine } Bd'
'           /bottomleft  { 2 1 HVLine } Bd'
'           /bottomright { 1 1 HVLine } Bd'
'           /lefttee     { 2 3 HVLine } Bd'
'           /righttee    { 1 3 HVLine } Bd'
'           /toptee      { 3 2 HVLine } Bd'
'           /bottomtee   { 3 1 HVLine } Bd'
'           /hline       { 3 0 HVLine } Bd'
'           /vline       { 0 3 HVLine } Bd'
'           /cross       { 3 3 HVLine } Bd'
'           /diamondsign {'
'				300 0 Mt 550 250 Lt 300 500 Lt 50 250 Lt closepath fill'
'           } Bd'
'       end'

'       /BuildGlyph {'
'           600 0'
'           0 -200 1000 1000'
'           setcachedevice'
'           exch dup /LineWidth get Slw /CharProcs get exch'
'           2 copy known not {pop /.notdef} if'
'           get exec'
'       } Bd'

'       /BuildChar {'
'           1 index /Encoding get exch get'
'           1 index /BuildGlyph get exec'
'       } Bd'

'       currentdict'
'   end'
'   definefont pop'
'} Bd'

;;; finds an Adobe font and converts it to use ISOLatin1Encoding
;;; usage: <isofontname> <fontname> DefISOLatin1Font
'/DefISOLatin1Font {'
'   findfont dup length dict begin'
'       { 1 index /FID ne {def} {pop pop} ifelse} forall'
'       /Encoding ISOLatin1Encoding def'
'       currentdict'
'   end'
'   definefont pop'
'} Bd'

;;; define shorthand for setting the various fonts
;;; sets font to standard font
'/Sf { /Courier-ISOLatin1 11 SF } Bd'
'/SfB { /Courier-ISOLatin1-Bold 11 SF } Bd'
'/SfI { /Courier-ISOLatin1-Oblique 11 SF } Bd'
'/SfBI { /Courier-ISOLatin1-BoldOblique 11 SF } Bd'
'/SfG { /VT100Graphics 11 SF } Bd'
'/SfGB { /VT100Graphics-Bold 11 SF } Bd'

;;; macro for underlining a string
;;; usage: <string> Ul
'/Ul {'
'   gsave dup stringwidth pop'
;;; move down 3/72" and draw a line under the string.
'   currentpoint -3 add newpath Mt 0 rlineto stroke grestore'
'} Bd'

;;; macro for drawing black outline around string
;;; usage: <string> <graylevel> Ol
'/Ol {'
'   gsave Sg dup stringwidth pop /swidth exch def'
'   currentpoint -3.75 add newpath Mt'
'   swidth 0 rlineto 0 11.5 rlineto'
'   swidth -1 mul 0 rlineto closepath fill grestore'
'} Bd'

'%%EndProlog'

'%%BeginSetup'

;;; define VT100 graphics fonts (standard/bold)

'%%BeginResource font VT100Graphics'
'/VT100Graphics 40 DefVT100GraphicsFont'
'%%EndResource'

'%%BeginResource font VT100Graphics-Bold'
'/VT100Graphics-Bold 70 DefVT100GraphicsFont'
'%%EndResource'

;;;; define ISOLatin1Encoding courier fonts (standard/bold/italic/bolditalic)

'%%BeginResource font Courier-ISOLatin1'
'/Courier-ISOLatin1 /Courier DefISOLatin1Font'
'%%EndResource'

'%%BeginResource font Courier-ISOLatin1-Bold'
'/Courier-ISOLatin1-Bold /Courier-Bold DefISOLatin1Font'
'%%EndResource'

'%%BeginResource font Courier-ISOLatin1-Oblique'
'/Courier-ISOLatin1-Oblique /Courier-Oblique DefISOLatin1Font'
'%%EndResource'

'%%BeginResource font Courier-ISOLatin1-BoldOblique'
'/Courier-ISOLatin1-BoldOblique /Courier-BoldOblique DefISOLatin1Font'
'%%EndResource'

'%%EndSetup'
;;; output text starts here...
}, explode <> identfn(%`\n`%)) |#);



define postscript_line_consumer(device);
	lvars
		device, consume,

		;;; document state
		done_preamble = false, done_endamble = false,
		line_count = 0, page_count = 0,
		longline_warning = false,
		control_char,

		;;; graphics state
		last_dattr, last_bold, last_ital, last_ul,
		last_inverse, last_light_text, last_graphics,

	  ;
	lconstant SPACE_CHARS = '\t\s\Ss\Sf\St\Sp\Sn';

	;;; forward decl;
	lconstant procedure Eject_page;

	lconstant macro (
		LINES_PER_PAGE = 66,
		Y_SPACING   = 11, ;;; spacing between each line, in .72"
		TOP_COORD   = (Y_SPACING + 1) * LINES_PER_PAGE,
		LEFT_MARGIN = 36, ;;; distance from left edge of page text in .72"
		MAX_COLUMNS = 80, ;;; most columns that can fit on a page
	  );

	;;; starting a new document
	define lconstant Output_preamble();
		npr('%!PS-Adobe-3.0');
		npr('%%Creator: Poplog (psprint.p driver JM/1/4/93)');
		npr('%%CreationDate: ' >< sys_convert_date(sys_real_time(),false));
		if device_open_name(device) then
			npr('%%Title: ' sys_>< device_open_name(device));
		endif;
		npr('%%For: ' >< sysgetusername(popusername));
		npr(preamble);
	enddefine;

	;;; reached the end of the document
	define lconstant Output_endamble();
		npr('%%Trailer');
		npr('\ndocsave restore\n');
		printf(page_count, '\n%%%%Pages: %p\n');
		npr('%%EOF');
	enddefine;

	;;; reached the end of the page
	define lconstant Output_end_page();
		;;; restore the saved context
		npr('\npgsave restore\n');
		;;; output this page
		npr('\nshowpage\n');
	enddefine;

	;;; starting a new page
	define lconstant Output_new_page();
		page_count + 1 -> page_count;
		printf(page_count, page_count, '\n%%%%Page: %p %p\n\n');
		;;; reset graphics state between each page to preserve page
		;;; independence.
		npr('%%BeginPageSetup');
		npr('/pgsave save def');
		npr('% set graphics state');
		npr('.5 Slw 0 Sg ');
		npr('%%EndPageSetup');
		false ->> last_dattr ->> last_graphics
			  ->> last_ul ->> last_inverse ->> last_light_text ->;
		undef ->> last_bold ->> last_ital ->;
	enddefine;

	;;; write characters in chars buffer
	lvars chars = [];
	define lconstant Flush_chars();
		if chars /== [] then
			pr('(');
			repeat #| explode(chars) |# times cucharout() endrepeat;
			pr(') ');
			if last_inverse then
				if last_light_text then
					pr('.8 Ol ')
				else
					pr('0 Ol ');
				endif;
			endif;
			if last_ul then pr('Ul '); endif;
			pr('Sh\n');
			sys_grbg_list(chars);
			[] -> chars;
		endif;
	enddefine;

	;;; write a single character to chars buffer
	define lconstant Write_char(c);
		lvars c;
		if last_graphics then
			c - #_< `\Gle` - `a` >_# -> c;
		else
			if c == `\^L` then
				Eject_page();
				true -> control_char;
				return;
			endif;
			if strmember(c, SPACE_CHARS) then
				;;; map tabs and other space characters to \s
				`\s` -> c;
			elseif strmember(c, '()\\') then
				;;; escape this character with a backslash
				conspair(`\\`, chars) -> chars;
			elseif c &&/=_0 128 then
				;;; its an eight-bit character - write it as an escaped octal
				conspair(`\\`, chars) -> chars;
				conspair((c // 64) + `0`, chars) -> chars -> c;
				conspair((c // 8) + `0`, chars) -> chars -> c;
				c + `0` -> c;
			endif;
		endif;
		conspair(c, chars) -> chars;
	enddefine;

	;;; set the current graphics state from a dchar and return it with
	;;; attributes stripped
	define lconstant Set_dchar_attributes(dchar) -> dchar;
		lvars dchar;
		lvars dattr, bold, ital, inverse, ul, light_text, color, graphics;
		dchar && 16:FF0000 -> dattr;
		dchar && 16:FF -> dchar;

		;;; see if its a special graphics character (see REF * ITEMISE)
		if dchar >= 129 and dchar <= 146 ->> graphics then
			;;; a couple of these characters can be handled in ISOLatin1
			if dchar == `\Go` then
				;;; degree
				8:260 -> dchar;
				false -> graphics;
			elseif dchar == `\G.` then
				;;; periodcentered
				8:267 -> dchar;
				false -> graphics;
			endif;
		endif;

		;;; if the character is a whitespace character, we ignore changes
		;;; to those attributes which aren't significant to whitespace (i.e.
		;;; bold and italic).
		if graphics == last_graphics and
				(if strmember(dchar, SPACE_CHARS) then
					last_dattr and (dattr && VEDCMODE_SP_SIG_BITS
									== last_dattr && VEDCMODE_SP_SIG_BITS)
				else
					dattr == last_dattr
				endif) then
			;;; ignore change
			return;
		endif;

		(dattr && VEDCMODE_COLOURNUM) >> VEDCMODE_COLOUR_SHIFT -> color;

		dattr &&/=_0 `\[u]` -> ul;
		dattr &&/=_0 `\[b]` -> bold;
		dattr &&/=_0 `\[i]` -> ital;
		color &&/=_0 1      -> inverse;
		not(color == 0 or color == 1) -> light_text;

		if bold /== last_bold or ital /== last_ital
				or graphics /== last_graphics then
			Flush_chars();
			;;; select the font.
			;;; note that there is not an italic version of the graphics font
			pr('Sf');
			if graphics then pr('G') endif;
			if bold then pr('B'); endif;
			if ital and not(graphics) then pr('I') endif;
			pr(' ');
		endif;

		if ul /== last_ul then
			Flush_chars();
		endif;

		if light_text /== last_light_text or last_inverse /== inverse then
			Flush_chars();
			pr(if light_text and not(inverse) then '.5'
				elseif light_text then '.2'
				elseif inverse then '1' else '0' endif);
			pr(' Sg ');
		endif;

		;;; record the graphics state
		inverse     -> last_inverse;
		light_text  -> last_light_text;
		ul          -> last_ul;
		bold        -> last_bold;
		ital        -> last_ital;
		graphics    -> last_graphics;
		dattr       -> last_dattr;
	enddefine;

	;;; output a single dchar
	define lconstant Output_dchar(dchar);
		lvars dchar;
		if control_char then
			Output_new_page();
			false -> control_char;
		endif;
		Write_char(Set_dchar_attributes(dchar));
	enddefine;

	;;; write a newline
	define lconstant Output_newline();
		Flush_chars();
		line_count + 1 -> line_count;
		if line_count == LINES_PER_PAGE then
			Output_end_page();
			0 -> line_count;
		endif;
	enddefine;

	;;; do a form feed
	define lconstant Eject_page();
		lconstant macro LAST_LINE = LINES_PER_PAGE - 1;
		;;; move to last line on page and then output a newline
		LAST_LINE -> line_count;
		Output_newline();
	enddefine;

	;;; write a single line
	define lconstant Output_line(dstring);
		lvars dstring;

		if line_count == 0 then ;;; starting a new page
			Output_new_page();
		endif;

		if dstring.datalength > MAX_COLUMNS then
			subdstring(1, MAX_COLUMNS, dstring) -> dstring;
			true -> longline_warning;
		endif;

		;;; skip empty lines
		false -> control_char;

		if dstring /= nullstring and skipchar(`\s`, 1, dstring) then
			;;; move to the next line
			printf(intof(TOP_COORD - (line_count * Y_SPACING)), LEFT_MARGIN,
				'\n%p %p Mt\n');
			appdata(dstring, Output_dchar); ;;; write the string
		endif;
		unless control_char then
			Output_newline();
		endunless;
	enddefine;

	unless isdevice(device) then
		syscreate(device, 1, false) -> device;
	endunless;

	discout(device) -> consume;

	;;; THE LINE CONSUMER (a lexical closure)
	procedure(line, device);
		lvars line, device;
		dlocal cucharout = consume, pr = syspr, pop_pr_quotes = false;
		unless done_preamble then
			Output_preamble();
			true -> done_preamble;
		endunless;
		if line == termin then
			unless done_endamble then
				unless line_count == 0 then
					Eject_page();
				endunless;
				Output_endamble();
				true -> done_endamble;
				if longline_warning then
					warning(0, 'PSPRINT: some lines were too long to fit on page.');
				endif;
				cucharout(termin);
			endunless;
		elseunless done_endamble then
			Output_line(line);
		endif;
	endprocedure(%device%);
enddefine;

endsection;


/* --- Revision History ---------------------------------------------------
--- Robert John Duncan, May 18 1995
		Changed procedure diamondsign for printing `\G#` to make it look
		like it does on screen. Removed procedures for printing `\Go` and
		`\G.` and mapped them to their ISOLatin1 equivalents instead.
 */
