/* --- Copyright University of Sussex 1990.  All rights reserved. ---------
 > File:           C.all/lib/lib/showcode.p
 > Purpose:        Tracing poplog virtual machine code planting procedures.
 > Author:         Jonathan Laventhol, Jan 30 1984 (see revisions)
 > Documentation:  HELP * SHOWCODE
 > Related Files:
 */

#_TERMIN_IF DEF POPC_COMPILING

;;; loading this file converts all the code planting procedures to print their
;;; name and their arguments, and makes sysnlabel make something printable.
;;; tracing is switched on and off by the boolean variable pop_show_code. the
;;; initial "sys" of the procedure names isn't printed, and calls to sysLABEL
;;; just print the argument, so sysSYNTAX(...); sysLABEL(sysnlabel) will print
;;; like this:
;;;
;;;     SYNTAX  hello procedure <false>
;;; label_3
;;;
;;; You can make the procedures do something other than print by redefining
;;; the variable procedure -popshowcdoepdr- (default is popshowcodesyspdr).
;;; The new procedure should take 2 arguments, the name of the code planting
;;; procedure (e.g. "sysSYNTAX" or "sysLABEL") and a vector of arguments.


section $-showcode => pop_show_code popshowcodepdr popshowcodesyspdr;

if identprops("make_code_showable") /== "undef" then
	[endsection;] -> proglist
endif;

global vars
	pop_show_code       =  false,
	procedure popshowcodepdr,
	;


define vars make_code_showable();

	;;; NB anyone updating the list should retain alphabetical order to
	;;; simplify checking for completeness (using ved_wordswith and ved_cdiff).

	applist([sysAND sysCALL sysCALLQ sysCALLS sysCOMPILE sysCONSTANT
			sysDLABEL sysDLVARS sysENDLBLOCK sysENDPROCEDURE sysERASE
			sysEXECUTE sysEXEC_COMPILE sysEXEC_OPTION_COMPILE
			sysFIELD sysGLOBAL sysGOTO sysGO_ON
			sysIDENT sysIFNOT sysIFSO sysLABEL sysLBLOCK sysLCONSTANT
			sysLOCAL sysLVARS
			;;; sysNEW_LABEL handled below
			sysNEW_LVAR sysOR sysPASSIGN sysPLOG_ARG_PUSH sysPLOG_IFNOT_ATOM
			sysPLOG_RESTART sysPLOG_RESTORE sysPLOG_SAVE sysPLOG_TERM_SWITCH
			sysPOP sysPROCEDURE sysPUSH sysPUSHQ sysPUSHS
			sysSWAP sysSYNTAX sysUCALL sysUCALLQ sysUCALLS
			sysUFIELD sysUPASSIGN sysVARS ],

		procedure(word);
			lvars procedure pdr, word;

			define lconstant trace_instr() with_props false;
				lvars args;
				if pop_show_code then
					consvector(pdnargs(pdr)) -> args;
					popshowcodepdr(word, args);
					explode(args)
				endif;
				pdr()
			enddefine;

			sysunprotect(word);
			valof(word) -> pdr;
			trace_instr -> valof(word);
			pdprops(pdr) -> pdprops(trace_instr);
			updater(pdr) -> updater(trace_instr);
			sysprotect(word)
		endprocedure);

	sysunprotect("sysNEW_LABEL");
	gensym(%"label_"%) -> valof("sysNEW_LABEL");
	"sysNEW_LABEL" -> pdprops(valof("sysNEW_LABEL"));
	sysprotect("sysNEW_LABEL");
	identfn -> make_code_showable
enddefine;


define global constant popshowcodesyspdr(name, args);
	lvars args, name;
	unless pop_syntax_only do
		unless name == "sysLABEL" or name == "sysDLABEL" do
			sp(4);
			pr(allbutfirst(3, name));
			sp(max(1, 8 - datalength(name) + 3));
		endunless;
		appdata(args, spr);
		cucharout(`\n`)
	endunless
enddefine;

popshowcodesyspdr -> popshowcodepdr;

make_code_showable();

endsection; /* showcode */

section;
vars showcode = true;	;;; for "uses"
endsection;


/*  --- Revision History ---------------------------------------------------
--- John Gibson, Jul  7 1990
		Added -sysEXEC_COMPILE-
--- Aaron Sloman, Jul  2 1990
		Declared variable showcode
--- John Williams, May 14 1990
		Replaced sys(U)(EXT)FIELD_VAL and sys(U)(EXT)SUBSCR with sys(U)FIELD
--- John Williams, Mar 26 1990
		No longer makes -sysPOP- the updater of -sysPUSH- etc.
--- Aaron Sloman, Mar 18 1990
	Added sysEXEC_OPTION_COMPILE sysEXT_SUBSCR sysSUBSCR sysUEXT_SUBSCR
	sysUSUBSCR, and re-arranged list into alphabetical order to simplify
	checking for completeness.
--- Ian Rogers, May 24 1989 - added sysEXT_FIELD_VAL sysUEXT_FIELD_VAL
--- John Gibson, Nov 23 1988
		Added sysLBLOCK, sysENDLBLOCK
--- Robert Duncan, Jul 13 1987 - added sysPLOG_ARG_PUSH
--- Simon Nichols, Feb 26 1987 - added sysPLOG_IFNOT_ATOM, sysPLOG_RESTART,
								 sysPLOG_TERM_SWITCH
--- John Williams, Feb 17 1987 - made updaters work again (see below!)
--- John Williams, Feb 10 1987 - removed 'chain' from 'trace_inst'
								 to make 'iscaller' etc work
--- John Gibson, Feb 4 1987 - removed 'updater' printing
--- John Williams, Jan 24 1987 - removed sysVAL, added sys(U)FIELD_VAL
--- John Williams, Nov 21 1986 - made it handle updaters
 */
