/* --- Copyright University of Sussex 1994. All rights reserved. ----------
 > File:            C.all/lib/proto/objectclass/auto/instance.p
 > Purpose:         Objectclass file
 > Author:          Steve Knight, HP Labs, 1992-1993 (see revisions)
 > Documentation:   HELP OBJECTCLASS
 > Related Files:
 */
;;; -- instance -------------------------------------------------------------
;;; syntax word for constructing instances in a convenient way.
;;;     instance CLASS [;]    ;;; this semi is optional
;;;         METHOD = EXPR;
;;;         METHOD = EXPR;
;;;         ...
;;;         METHOD = EXPR[;]        ;;; last semi is optional
;;;     endinstance
;;;
;;; define-extension for constructing instances in a convenient way.
;;;     define :instance [IDENTSPEC] VAR:CLASS [;]
;;;         METHOD = EXPR;
;;;         METHOD = EXPR;
;;;         ...
;;;         METHOD = EXPR[;]
;;;     enddefine

section $-objectclass =>
	define_instance
	instance
	endinstance;

define lconstant procedure process_method_init( Lvar, closer );
	lvars Lvar, closer;
	lvars name = itemread();
	unless name.isword do
		mishap( 'OBJECT: invalid method name', [^name] )
	endunless;
	pop11_need_nextitem( "=" ).erase;
	pop11_comp_expr();
	sysPUSH( Lvar );
	sysUCALL( name );
	unless
		pop11_try_nextreaditem( ";" ) or
		null( proglist ) or
		hd( proglist ) == closer
	do
		;;; This will mishap with the right message.
		pop11_need_nextreaditem( closer ).erase;
	endunless;
enddefine;

define lconstant procedure bindings( Lvar, closer ); lvars Lvar, closer;
	until pop11_try_nextreaditem( closer ) do
		process_method_init( Lvar, closer );
	enduntil;
enddefine;

global vars syntax endinstance;

define lconstant procedure instance_to( closer ); lvars closer;
	dlocal pop_new_lvar_list;

	lvars class = readword();
	lvars Lvar = sysNEW_LVAR();

	sysPUSH( class <> "_key" );
	sysCALLQ( class_new );
	sysCALLS( undef );
	sysPOP( Lvar );

	;;; optional semi-colon.
	pop11_try_nextreaditem( ";" ).erase;

	bindings( Lvar, closer );
	sysPUSH( Lvar );
enddefine;

define global syntax instance;
	instance_to( "endinstance" )
enddefine;


;;; Test for the presence of VED.
if testdef vedprocess then
	unless lmember( "instance", vedopeners ) do
		[instance ^^vedopeners] -> vedopeners;
		[endinstance ^^vedclosers] -> vedclosers;
	endunless;
endif;

define lconstant procedure define_instance_to( closer ); lvars closer;
	dlocal pop_new_lvar_list;

	lvars idspec = read_identspec( "vars" );
	lvars Lvar = readword();
	pop11_need_nextreaditem( ":" ).erase;

	identspec_declare( idspec )( Lvar, 0 );
	lvars class = readword();

	sysPUSH( class <> "_key" );
	sysCALLQ( class_new );
	sysCALLS( undef );
	sysPOP( Lvar );

	;;; optional semi-colon.
	pop11_try_nextreaditem( ";" ).erase;

	bindings( Lvar, closer );
enddefine;

define :define_form global instance;
	define_instance_to( "enddefine" )
enddefine;

endsection;

/* --- Revision History ---------------------------------------------------
--- Integral Solutions Ltd (Julian Clinton), Feb 16 1994
		Replaced readitem with itemread to allow macro expansion
		within "instance...endinstance" in -process_method_init-.
 */
