/* --- Copyright University of Sussex 1993. All rights reserved. ----------
 > File:            C.all/lib/proto/objectclass/src/globals.p
 > Purpose:         Objectclass file
 > Author:          Steve Knight, HP Labs, 1992-1993
 > Documentation:   HELP OBJECTCLASS
 > Related Files:
 */
;;; -- Global variables for the Objectclass package. ------------------------

compile_mode :pop11 +strict;

section $-objectclass;

;;; This variable is used to optimise unlinking.
;;;
vars all_classed_methods_are_unlinked = true;

;;; This variable indicates whether or not it is safe to trash method
;;; tables when we optimise methods.  It is set to false, for example,
;;; by loading -method_path-.
;;;
vars can_trash_method_table = true;

;;; This variable is used to extend the compilation process.  This allows
;;; us to make method tracing autoloadable.
vars relink_hook = erase;


;;; -- invokes_call_next_method
;;; Flag set on method parts to allow the method linker to generate
;;; suitable code for parts that use call_next_method.
constant procedure invokes_call_next_method =
	newproperty( [], 20, false, "tmparg" );
constant procedure c_invoke_call_next_method =
	procedure() with_props cICNM; internal_error() endprocedure;
constant procedure u_invoke_call_next_method =
	procedure() with_props uICNM; internal_error() endprocedure;
CALL_MODE -> c_invoke_call_next_method.invokes_call_next_method;
UCALL_MODE -> u_invoke_call_next_method.invokes_call_next_method;



;;; -- isobsolete
;;; Table flagging obsolete classes.
constant procedure isobsolete = newproperty( [], 8, false, "tmparg" );


;;; -- class_extern_posn
;;; Table describing where the >-> is inside a class.

constant procedure class_extern_posn =
	newanyproperty(
		[], 20, 1, false,
		false, false, "tmparg",
		false, false
	);

;;; -- fields_of_class
;;; Table describing the format of a class - similar to key_of_dataword.
;;; Returns a list of slots.

constant procedure fields_of_class =
	newanyproperty(
		[], 20, 1, false,
		false, false, "tmparg",
		false, false
	);

;;; -- Supers of Class ------------------------------------------------------

;;; Table describing the "super" relationship.  Note that it *must* be
;;; temporary to avoid store clogging up during recompilation.
;;; The supers are stored in order of dominance ie. dominant classes come
;;; early in the list.
;;;
constant procedure supers_of_class =
	newanyproperty(
		[], 20, 1, false,
		false, false, "tmparg",
		[], false
	);

;;; iterate upwards over the inheritance-net, in inheritance order, over the
;;; supers of k, applying p at each point.
define appsupers( k, p ); lvars k, p;
	lvars procedure seen = newproperty( [], 20, false, "perm" );

	define lconstant procedure Scan( here ); lvars here;
		unless seen( here ) do
			true -> seen( here );
			p( here );
			applist( supers_of_class( here ), Scan )
		endunless
	enddefine;

	Scan( k )
enddefine;

define all_supers_of_class( k ); lvars k;
	[%
		applist(
			k.supers_of_class,
			procedure( k ); lvars k;
				appsupers( k, identfn )
			endprocedure
		)
	%]
enddefine;


;;; -- Inferiors of class ---------------------------------------------------
;;; At some points we need the inverse to supers -- namely inferiors.

constant procedure infs_of_class =
	newanyproperty(
		[], 20, 1, false,
		false, false, "tmparg",
		false, false
	);

;;; In the case of extantclasses, we need to be able to flush the properties
;;; for the key.  This guards against recompilation of the extantclass with
;;; altered "isa" values.  The other tables don't matter because they get
;;; directly updated.
;;;
define clear_infs( key ); lvars key;
	lvars c;
	for c in supers_of_class( key ) do
		lvars children = infs_of_class( c );
		false -> children( key );                ;;; disinherit key
	endfor;
enddefine;

define lconstant procedure New_inf_map();
	newproperty( [], 4, false, "tmpboth" )
enddefine;

define add_inf( x, y ); lvars x, y;
	lvars t = infs_of_class( y );
	unless t then
		New_inf_map() ->> t -> infs_of_class( y )
	endunless;
	true -> t( x )
enddefine;

define dump_infs_of_class( c ); lvars c;
	lvars t = infs_of_class( c );
	if t then
		fast_appproperty( t, erase )
	endif
enddefine;

;;; Returns all the inferiors of the keys (list) supplied.  All duplicates
;;; are removed.  Used by relink_method.  Leaves the results on the stack.
;;;
define all_infs( keys ); lvars keys;
	lvars isseen = newproperty( [], 20, false, "perm" );
	until null( keys ) do
		lvars k = keys.dest -> keys;
		unless isseen( k ) do
			true -> isseen( k );
			[% dump_infs_of_class( k ) % ^^keys] -> keys;
			k;
		endunless;
	enduntil;
enddefine;

endsection;
