/* --- Copyright University of Sussex 1993. All rights reserved. ----------
 > File:            C.all/lib/proto/objectclass/src/interpreter.p
 > Purpose:         Objectclass file
 > Author:          Steve Knight, HP Labs, 1992-1993
 > Documentation:   HELP OBJECTCLASS
 > Related Files:
 */
compile_mode :pop11 +strict;

section $-objectclass;

;;; -- Executing a Tree -----------------------------------------------------

;;; It is important that using this dosn't build up lots of unneeded
;;; activation records.  As a consequence, we chain whenever possible.
;;; This won't save much, though.

constant procedure run_tree;

define run_next( L, level, call ); lvars level, L, call;

	;;; This check ensures we don't POP from an empty stack!
	returnif( L == [] );

	lvars k = datakey( subscr_stack( level ) );
	repeat
		lvars b = destpair( L ) -> L;           ;;; fast_destpair OK.
		if lmember( k, keysBranch( b ) ) then   ;;; fast_lmember OK.
			lvars TS = treesBranch( b );
			until TS == [] do
				lvars t = TS.destpair -> TS;    ;;; fast_destpair OK.
				if L == [] and TS == [] then
					fast_chain( t, level, call, run_tree )
				else
					run_tree( t, level, call )
				endif
			enduntil;
		endif;
		quitif( L == [] );
	endrepeat;
enddefine;

define run_tree( T, level, call ); lvars T, level, call;
	level + 1 -> level;                         ;;; fi_+ OK
	lvars d = T.defaultTree;
	lvars a = T.actionTree;
	lvars n = T.nextTree;

	if a then
		run_next( n, level, call );
		if d then
			run_tree( d, level, call )
		endif;
		chain( a, call )                        ;;; fast_chain OK.
	elseif d then
		run_next( n, level, call );
		if d then
			fast_chain( d, level, call, run_tree )
		endif
	else
		fast_chain( n, level, call, run_next )
	endif;
enddefine;


;;; -- Creating, Linking and Unlinking the interpreter ----------------------

define c_interpret( T ); lvars T;
	if T then run_tree( T, 0, call_method_part ) endif;
	call_method_failure();
enddefine;

define u_interpret( T ); lvars T;
	if T then run_tree( T, 0, call_method_part ) endif;
	() -> call_method_failure();
enddefine;

define relink_interpreter( Interpreter, Generic ); lvars Generic, Interpreter;
	lvars MT = Generic.method_table;

	lockMethodTable( MT );

	lvars centries = cEntriesMethodTable( MT );
	lvars uentries = uEntriesMethodTable( MT );

	lvars need_updater = not( null( uentries ) );

	lvars tree  = net_to_tree( centries );
	lvars utree = need_updater and net_to_tree( uentries );

	unlockMethodTable( MT );

	lvars carity = cArityMethodTable( MT );
	lvars uarity = need_updater and uArityMethodTable( MT );

	if tree then
		instantiate( tree, false )
	endif;
	if need_updater and utree then
		instantiate( utree, false )
	endif;

	c_interpret -> Interpreter.indirect_pdpart;
	tree -> indirect_frozval( 1, Interpreter );
	set_pdnargs( carity, Interpreter );

	if need_updater then
		u_interpret(% utree %) -> Interpreter.updater;
		set_pdnargs( uarity, Interpreter.updater );
	else
		false -> Interpreter.updater
	endif;
enddefine;

define unlink_interpreter( Int, Gen ); lvars Int, Gen;
	lvars IG = conspair( Int, Gen );
	IG -> indirect_frozval( 1, Int );
	procedure( IntGen ); lvars IntGen;
		relink_interpreter( IntGen.destpair );
		chain( IntGen.front );
	endprocedure -> Int.indirect_pdpart;
	procedure( IntGen ); lvars IntGen;
		relink_interpreter( IntGen.destpair );
		lvars i = IntGen.front;
		if i.updater then
			chain( i );
		else
			() -> i();      ;;; Generate the right error message.
		endif
	endprocedure(% IG %) -> Int.indirect_updater;
enddefine;

define new_interpreter( Gen ) -> Int; lvars Int, Gen;
	newindirect_n( identfn, undef, 1 ) -> Int;
	unlink_interpreter( Int, Gen );
enddefine;

endsection;
