/* --- Copyright University of Sussex 1993. All rights reserved. ----------
 > File:            C.all/lib/proto/objectclass/src/method_form.p
 > Purpose:         Objectclass file
 > Author:          Steve Knight, HP Labs, 1992-1993
 > Documentation:   HELP OBJECTCLASS
 > Related Files:
 */
;;; -- Method definition syntax ---------------------------------------------

section $-objectclass;


;;; The define-hook -- sort of user defineable.  Agreed to put this in after
;;; a conversation with Jonathan Cunningham.  Definitely a good idea.
;;; Just a stub at present.
define vars procedure define_hook( keyword, name ); lvars keyword, name;
enddefine;

;;; Dynamically bound variable holding the method name & call/update mode.
vars MethodName = false;
vars MethodMode = false;

;;; These variables holds the list of input locals, their classes and
;;; the output variables for the currently compiling method.
vars InLocals = false;
vars InClasses = false;
vars OutLocals = false;

;;; This variable is dlocalised to a property-table.  This allows various
;;; internal constructs to stash information for later use.
vars CurrentProperties = false;


;;; -- define :method -------------------------------------------------------
;;; define :<method> [updaterof] foo( x[:C], y[:C] ) -> r1 -> ... -> rN;
;;;     ...
;;; enddefine;

define make_shared_slot( initial ) -> Class_slot;
	lvars initial;
	lconstant procedure Class_slot;

	lvars value = conspair( false, initial );

	define lconstant procedure Class_slot( x ); lvars x;
		if fast_front( value ) then
			fast_back( value )
		else
			cons_with check_one {% fast_back( value )( x ) %} ->> fast_back( value );
			true -> fast_front( value );
		endif
	enddefine;

	define updaterof Class_slot() with_nargs 2;
		erase();
		-> fast_back( value );
		true -> fast_front( value );
	enddefine;
enddefine;

define lconstant procedure Make_slot( initial ); lvars initial;
	newanyproperty(
		[], 20, 1, false,
		false, false, "tmparg",
		false,
		procedure( k, p ); lvars k, p;
			cons_with check_one {% initial( k ) %} ->> p( k )
		endprocedure
	)
enddefine;

;;; Note the use of sysPUSHQ because the classes can be false.
;;;
define deref_classes( iclasses ); lvars iclasses;
	[%
		lvars i;
		for i in iclasses do
			if i then currentval else identfn endif( i );
		endfor;
	%]
enddefine;

;;; ilocals is a list of the input locals (words)
;;; iclasses is a list of the classes associated with the input locals (words)
define procedure get_ilocals() -> ( idecs, ilocals, iclasses ); lvars idecs, ilocals, iclasses;
	pop11_need_nextreaditem( "(" ).erase;
	lvars ilocals = [], iclasses = [], idecs = [];
	until pop11_try_nextreaditem( ")" ) do
		pop11_try_nextreaditem( "," ).erase;   ;;; sloppy way to do optional ","
		lvars ( D, L ) = read_local();
		[ ^D ^^idecs ] -> idecs;
		[ ^L ^^ilocals ] -> ilocals;
		if pop11_try_nextreaditem( ":" ) then
			lvars C = readitem();
			unless C.isword do
				mishap( 'DEFINE_METHOD: class must be a word', [^C] )
			endunless;
			lvars kname = derive_key_name( C );
			lvars k = currentval( kname );
			if k.isclass then
				kname :: iclasses -> iclasses;
			else
				mishap( 'CLASS NAME REQUIRED', [^kname] )
			endif
		else
			false :: iclasses -> iclasses
		endif;
	enduntil;
enddefine;

define procedure check_ilocals( ilocals, iclasses, keyword );
	lvars ilocals, iclasses, keyword;
	if keyword == "IF_NEEDED" or keyword == "SHARED_SLOT" then
		unless ilocals.length == 1 and iclasses.length == 1 do
			mishap( 'One and only one typed input local allowed', [] )
		endunless
	endif;
enddefine;

;;; This procedure synthesises a meaningful name for a method part.
;;; It is passed the initial segment of proglist which
;;; MUST be terminated by a ";" -- and this is guaranteed by the point of
;;; call of -Synth_method_name-.  The name generated is helpful when using
;;; lib debugger.
;;;
;;; Note that dlocalising "identfn" is completely wrong -- however it
;;; is the wrong place to correct this dumb mistake which has been
;;; perpetuated through laziness (in my view).
;;;
define lconstant procedure Synth_method_name( plist ); lvars plist;
	dlocal cucharout = identfn;
	consstring(#|
		lvars item;
		for item in plist do
			quitif( item == ";" );
			if item == ")" then
				pr( item );
				quitloop
			endif;
			if item == "updaterof" then
				pr( item );
				cucharout( `_` );
			else
				pr( item )
			endif
		endfor;
	|#)
enddefine;

define lconstant procedure Make_action( idecs, ilocals, odecs, olocals, nargs, props, def_props, decl, closing_keyword );
	lvars ilocals, olocals, idecs, odecs, nargs, props, closing_keyword, decl, def_props;
	sysPROCEDURE(
		which_pdprops( props, def_props, decl ),
		nargs or ilocals.length
	);
	applists_n( ilocals, idecs, 2, apply );
	applists_n( olocals, odecs, 2, apply );
	applist( ilocals, sysPOP );
	pop11_comp_stmnt_seq_to( closing_keyword ).erase;
	sysLABEL( "return" );
	applist( olocals, sysPUSH );
	sysENDPROCEDURE()
enddefine;

;;; Note these keywords are used by autoloadable define:forms.
define lconstant procedure Cnvt_action( keyword, action ); lvars keyword, action;
	if keyword == "IF_NEEDED" then
		Make_slot
	elseif keyword == "SHARED_SLOT" then
		make_shared_slot
	else
		identfn
	endif( action )
enddefine;

define do_method_form ( MethodName, MethodMode, decl, def_props, closing_keyword );
	dlocal MethodName, MethodMode, InLocals, InClasses, OutLocals;
	lvars idecs, action, odecs, def_props, decl, closing_keyword;
	get_ilocals() -> ( idecs, InLocals, InClasses );
	read_olocals() -> ( odecs, OutLocals );
	lvars ( nargs, props ) = read_nargs_and_props();
	pop11_need_nextreaditem( ";" ).erase;

	dlocal CurrentProperties = newproperty( [], 8, false, "perm" );

	mode_identspec_declare( decl, MethodMode )( MethodName, "procedure" );
	bind_method( MethodName, MethodMode );

	lvars action = Make_action( idecs, InLocals, odecs, OutLocals, nargs, props, def_props, decl, closing_keyword );

	if action.isclosure then
		;;; Boo, hiss!  We have captured lexical variables. We can't
		;;; cope with this properly.
		mishap( 'METHOD CAPTURING LEXICAL VARIABLE (stupid restriction, sorry)', [^MethodName] )
	endif;

	if CurrentProperties( "invokes_call_next_method" ) then
		MethodMode -> invokes_call_next_method( action )
	endif;
	return( action, InLocals, deref_classes( InClasses ) );
enddefine;

define method_form( keyword, closing_keyword ); lvars keyword, closing_keyword;
	dlocal MethodName, MethodMode;
	lvars idecs, odecs;
	pop11_try_nextreaditem( "updaterof" ).flag_to_mode -> MethodMode;
	lvars decl = read_identspec( default_default_idspec() );
	lvars plist = proglist;     ;;; cache proglist for synthesising names.
	if MethodMode == UCALL_MODE then
		"updaterof" :: plist -> plist;
	endif;
	readitem() -> MethodName;
	define_hook( keyword, MethodName );
	if pop11_try_nextreaditem( ";" ) then
		mishap( 'OBSOLETE SYNTAX - use define :generic', [] );
	else
		lvars ( action, in_locals, in_classes ) =
			do_method_form( MethodName, MethodMode, decl, Synth_method_name( plist ), closing_keyword );

		check_ilocals( in_locals, in_classes, keyword );
		Cnvt_action( keyword, action ) -> action;

		update_method(
			action,
			MethodMode,
			in_classes,
			currentval( MethodName )
		)
	endif
enddefine;

endsection;

;;; -------------------------------------------------------------------------
;;; Modified 4/10/92, sfk
;;;     *   Ensured that MethodName was dynamically bound to method_form.
;;;         This wasn't correct previously and the error message for
;;;         using call_next_method outside of a procedure was
;;;         inevitably incorrect.
;;;     *   Cleaned up spelling errors inside comments.
;;; -------------------------------------------------------------------------
;;; Modified, 6/12/92, sfk
;;;     *   Cleaned up declaration of :define_form method.
;;; -------------------------------------------------------------------------
;;; Modified, 15/04/93, sfk
;;;     *   Changed spurious "GENERIC" prefix on error messages
;;; -------------------------------------------------------------------------
;;; Modified, 20/04/93, sfk
;;;     *   Uses the read_olocals utility, shared by the define:pd form.
;;; -------------------------------------------------------------------------
;;; Modified, 30/05/93, sfk
;;;     *   Extensively rewritten to reflect the new compilation
;;;         strategy for classes.
;;; -------------------------------------------------------------------
;;; Modified, 4/6/93, sfk
;;;     *   Changed the mode value for invokes_call_next_method to use
;;;         CALL_MODE and UCALL_MODE.
;;; -------------------------------------------------------------------------
;;; Modified, 31/8/93, sfk
;;;     *   Removed the define :method <name>; syntax.  Replaced
;;;         by define :generic <name>.
;;;     *   Made define_method autoloadable.  This is a fairly
;;;         irrelevant change that I made for symmetry.  It is sensible
;;;         in principle -- but makes no change in practice.
;;; -------------------------------------------------------------------------
