/* --- Copyright University of Sussex 1993.  All rights reserved. ---------
 > File:		   C.all/lib/ved/ved_jj.p
 > Purpose:		   Full text justification (both left and right margins)
 > Author:		   Chris Slymon, June 1983 (see revisions)
 > Documentation:  HELP * FORMAT
 */
compile_mode :pop11 +strict;

/*	From an original design by Dave Roberts.
	Any line that is followed by a line beginning with a space is assumed to
	be the last line of a paragraph and is not adjusted.
*/

section;

define vars ved_jj;
	lvars	leftmcol, string, place, num, oldmax, mustrefresh,
			oldline = vedline,
			oldoffset = vedlineoffset,
			oldchanged = vedchanged,
			wasediting = vedediting;

	dlocal	vedstatic = false, vedediting = false, vedautowrite = false;

	define lconstant wordright();
		;;; Move right one word, delimited by space

		lvars string = vedthisline(), col;

		locchar(`\s`,vedcolumn,string)-> col;
		if col and col fi_<= vvedlinesize then
			col -> vedcolumn;
			if skipchar(`\s`,vedcolumn,string) ->> col then
				col
			endif;
		else
			vvedlinesize fi_+ 1
		endif -> vedcolumn;
		;;; return true unless reached end
		vedcolumn fi_< vvedlinesize
	enddefine;

	define lconstant insertspaces();
		;;; Pad current line with spaces
		lvars	gaps = 0, count, spaces = vedlinemax fi_- vvedlinesize,
				quotient, remainder, done = 0, c;

		1 -> vedcolumn;
		returnif((vedcurrentchar() ->> c) == `\Sf` or c == `\Sp`
					or spaces == 0);

		vedtextleft();
		;;; count the gaps
		while wordright() do gaps fi_+ 1 -> gaps endwhile;
		returnif(gaps == 0);		;;; nowhere to pad

		spaces // gaps -> quotient -> remainder;

		vedtextleft();

		while wordright() do
			fast_for count to quotient do vedcharinsert(`\s`) endfor;
			spaces fi_- quotient -> spaces;
			unless remainder == 0 then
				done fi_+ remainder -> done;
				if done fi_>= gaps then
					vedcharinsert(`\s`);
					done fi_- gaps -> done
				endif
			endunless;
			quitif(spaces == 0) 	;;; done
		endwhile
	enddefine;

	define lconstant vedwindowbottom();
		vedlineoffset + vedwindowlength - 1
	enddefine;

	false -> vedchanged;
	not(vvedmarkhi <= vedlineoffset or vvedmarklo > vedwindowbottom())
		-> mustrefresh;

	if isprocedure(vedleftmargin) then vedleftmargin()
	else vedleftmargin
	endif fi_+ 1 -> leftmcol;

	vvedmarkhi -> oldmax;

	vedpositionpush();
	ved_gobble();
	ved_fill();
	vedmarkfind();

	while vedline fi_< vvedmarkhi do
		;;; Do not fill last line of paragraph, or line without gaps
		fast_subscrv(vedline fi_+ 1, vedbuffer) -> string;
		unless datalength(string) fi_< leftmcol
		or strmember(subscrs(leftmcol, string), '\s\t')
		or strmember(subscrs(1, string), '\Sf\Sp')
		then
			insertspaces()
		endunless;
		vednextline()
	endwhile;

	vedpositionpop();

	wasediting -> vedediting;

	oldmax - vvedmarkhi -> num;
	if mustrefresh then
		if vvedmarkhi <= vedlineoffset then
			;;; range just disappeared off the top of screen
			oldmax - oldoffset -> num; ;;; lines disappeared
			2 -> place;
		elseif vvedmarklo <= vedlineoffset then
			;;; Start of range off screen, part visible.
			;;; calculate reduction
			if vvedmarkhi < vedwindowbottom() then
				;;; Number of lines reduced in range
				(oldmax - oldoffset) - (vvedmarkhi - vedlineoffset) -> num;
				if num > 0 then vvedmarkhi - vedlineoffset + 1
				else 2
				endif -> place;	;;; from where to scroll up or down
			else
				0 -> num;
			endif;
		else
			;;; start of range on screen
			if vvedmarkhi >= vedwindowbottom() then
				;;; Start of range on screen, extends to bottom
				0 -> num;	;;; no scrolling needed
			else
				oldmax - vvedmarkhi -> num;	;;; lines shortened
				if num > 0 then vvedmarkhi else vvedmarklo endif
					- vedlineoffset + 1 -> place;
			endif;
		endif;
		;;; now do refreshing and scrolling
		if num > 0 then
			;;; scroll up and refresh bottom
			repeat num times
				vedscreenscrollregionup(place, vedwindowlength);
			endrepeat;
			vedrefreshrange(
				vedwindowbottom() - num + 1,
				vedwindowbottom(), undef);
		elseif num < 0 then
			repeat -num times
				vedscreenscrollregiondown(place, vedwindowlength)
			endrepeat;
		else
			;;; num = 0, so no scrolling
		endif;
		vedrefreshrange(vvedmarklo,vvedmarkhi,undef);
	endif;
	if oldchanged then oldchanged + 1 else 1 endif -> vedchanged;
enddefine;

endsection;


/*	--- Revision History ---------------------------------------------------
--- John Gibson, Apr 22 1993
		Made insertspaces ignore lines beginning with \Sf or \Sp
--- Aaron Sloman, May 29 1990
	Altered to minimize screen activity - complained about in bug report
--- Aaron Sloman, Jan  3 1987 speeded up substantially. Changed to take note
	of leading tabs as well as leading spaces. Uses only integers -- no ratios
	and need not round.
	vedautowrite set false
--- Mark Rubinstein, Jan 23 1986 - fixed to ensure that if vedleftmargin is a
	procedure then its value is got first.
 */
