;;; --- Copyright University of Sussex 1990. All rights reserved. ----------
;;; File:           C.all/lisp/flib/f_C.p
;;; Purpose:        FORMAT_PRINT - PRINTING CHARACTERS  ~C
;;; Author:         John Williams, Dec  9 1985 (see revisions)
;;; Documentation:  HELP * FORMAT_PRINT
;;; Related Files:  LIB * FORMAT_PRINT

uses format_print;

section $-lisp;

global vars print_escape print_pretty;
global vars procedure lisp_pr;
global constant procedure char_name;

section fpr;

procedure();
	lvars char lchar name;
	dlocal print_escape print_pretty;
	next_f_arg() -> char;
	if f_subsystem == "lisp" then
		char -> lchar;
		destcharacter(lchar) -> char
	else
		false -> lchar;
		unless isinteger(char) and char fi_>= 0 and char fi_< 128 do
			mishap(char, 1, 'Invalid ASCII Code')
		endunless;
	endif;
	if f_colon then
		;;; PRINT CHARACTER NAMES
		if lchar then
			char_name(lchar)
		elseif char fi_< `\s` then
			'Control  ' ->> name;
			char fi_+ 64 -> fast_subscrs(9, name)
		elseif char == `\s` then
			'Space'
		elseif char == `\^?` then
			'Delete'
		else
			false
		endif -> name;
		if pop_true(name) then
			appdata(name, cucharout)
		else
			cucharout(`"`);
			cucharout(char);
			cucharout(`"`)
		endif;
		if f_at then
			;;; INDICATE KEYS
			if char ==  `\b` then
				' (the BS key)'
			elseif char ==  `\t` then
				' (the TAB key)'
			elseif char ==  `\n` then
				' (the LF key)'
			elseif char ==  `\r` then
				' (the RETURN key)'
			elseif char ==  `\^[`then
				' (the ESC key)'
			elseif char ==  `\^?`then
				' (the DEL key)'
			elseif lchar and char fi_< `\s` then
				' (Control  )' ->> name;
				char fi_+ 64 -> fast_subscrs(11, name)
			else
				false
			endif -> name;
			if name then
				appdata(name, cucharout)
			endif
		endif
	elseif f_at then
		;;; OUTPUT SHOULD BE RE-READABLE
		if lchar then
			true -> print_escape;
			nil -> print_pretty;
			lisp_pr(lchar)
		else
			cucharout(```);
			if char fi_<= `\s` then
				cucharout(`\\`);
				if     char ==  `\b` then `b`
				elseif char ==  `\t` then `t`
				elseif char ==  `\n` then `n`
				elseif char ==  `\r` then `r`
				elseif char ==  `\s` then `s`
				elseif char ==  `\\` then `\\`
				else cucharout(`^`), char fi_+ 64
				endif
			elseif char == `\^?` then
				cucharout(`\\`);
				cucharout(`^`);
				`?`
			else
				char
			endif;
			cucharout();
			cucharout(```)
		endif
	else
		;;; JUST OUTPUT CHAR
		cucharout(char)
	endif
endprocedure -> f_proc(`C`);

endsection;
endsection;


;;; --- Revision History ---------------------------------------------------
;;; --- John Williams, Jul 19 1990
;;; 		Changed -char_int- to -destcharacter-
