#| --- Copyright University of Sussex 1994. All rights reserved. ----------
 | File:            C.all/lisp/modules/profile.lsp
 | Purpose:         Lisp interface to Pop-11 LIB * PROFILE.P
 | Author:          John Williams, Mar  4 1992 (see revisions)
 | Documentation:   HELP * PROFILE
 | Related Files:   LIB * PROFILE.P
 |#

(cl:provide :profile)

(cl:in-package :poplog)

(export '(profile
			*profile-exclude*
			*profile-excluded-packages*
			*profile-gc-trace*
			*profile-interval*
			*profile-lisp-only*
			*profile-output*
			*profile-show-max* ))

(pop11)


section $-lisp;

lisp_compile_mode;

uses profile;

vars
	profile_excluded_packages   =   [],
	profile_lisp_only           =   true,
	profile_output              =   make_synonym_stream(@*TRACE-OUTPUT*),
	;

vars pop11_profile_apply;
if isundef(pop11_profile_apply) then
	profile_apply -> pop11_profile_apply
endif;


define global vars lisp_profile_include(pdr) -> name;
	lvars pdr, name;
	f_name(pdr) -> name;
	if (profile_lisp_only and not(issymbol(name)))
	or (fast_member(name, profile_exclude))
	or (issymbol(name)
		and fast_member(symbol_package(name), profile_excluded_packages))
	then
		false -> name
	endif
enddefine;


define global vars lisp_profile_apply() with_nargs 1;
	dlocal
		profile_include     =   lisp_profile_include,
		pr                  =   lisp_pr,
		standard_output     =   profile_output,
		;
	pop11_profile_apply()
enddefine;


define Profile(forms, nresults);
	lvars forms, nresults;
	sysPROCEDURE(false, 0);
	Progn(forms, nresults);
	sysENDPROCEDURE();
	sysPUSHQ();
	sysCALL("ident lisp_profile_apply")
enddefine;


lisp_export(Profile, @PROFILE, "special");

vars active (
	pf_gc_trace     =   boolean_variable(% ident profile_gc_trace %),
	pf_lisp_only    =   boolean_variable(% ident profile_lisp_only %),
	);

lispsynonym(@*PROFILE-EXCLUDE*,             "profile_exclude");
lispsynonym(@*PROFILE-EXCLUDED-PACKAGES*,   "profile_excluded_packages");
lispsynonym(@*PROFILE-GC-TRACE*,            "pf_gc_trace");
lispsynonym(@*PROFILE-INTERVAL*,            "profile_interval");
lispsynonym(@*PROFILE-LISP-ONLY*,           "pf_lisp_only");
lispsynonym(@*PROFILE-OUTPUT*,              "profile_output");
lispsynonym(@*PROFILE-SHOW-MAX*,            "profile_show_max");


endsection;


/* Back to Lisp for the revision history */

lisp

#| --- Revision History ---------------------------------------------------
--- John Williams, Sep 21 1994
		Replaced WID with new ident syntax.
 |#
