/* --- Copyright University of Sussex 1995. All rights reserved. ----------
 > File:            C.all/lisp/src/pprint.p
 > Purpose:         Common Lisp pretty printer
 > Author:          John Williams, Feb 25 1986 (see revisions)
 > Documentation:
 > Related Files:   C.all/lisp/src/print.p
 */

lisp_compile_mode;

section $-lisp;

/* The *PRINT-CIRCLE* flag */

vars pc_lablist = [];

define lconstant Pc_label() with_nargs 1;
	list_assoc(pc_lablist)
enddefine;


define updaterof lconstant Pc_label(lab, item);
	lvars item lab;
	acons(item, lab, pc_lablist) -> pc_lablist
enddefine;


define lconstant Pc_mark() with_nargs 1;
	lconstant Printed = newproperty([], 32, false, false);
	dlvars Labcount = 0;
	dlocal
		cucharout                       =   erase,
		standard_output                 =   Null_io,
		0 % clearproperty(Printed), clearproperty(Printed) %,
		;

	define dlocal _lisppr(item);
		lvars item lab;
		if issimple(item)
		or ischaracter(item)
		or (issymbol(item) and symbol_package(item) /== nil) then
			return
		elseif Printed(item) then
			unless Pc_label(item) do
				Labcount fi_- 1 ->> Labcount -> Pc_label(item)
			endunless
		else
			true -> Printed(item);
			lispsyspr(item)
		endif
	enddefine;

	_lisppr();
enddefine;


define lconstant Pc_printed(item);
	lvars item l;
	if (Pc_label(item) ->> l) then
		cucharout(`#`);
		if l fi_< 0 then
			sys_syspr(negate(l) ->> Pc_label(item));
			cucharout(`=`);
			false
		else
			sys_syspr(l);
			cucharout(`#`);
			true
		endif
	else
		false
	endif
enddefine;


define lconstant Pc_lisppr(item);
	lvars item;
	unless Pc_printed(item) do
		lispsyspr(item)
	endunless
enddefine;


/* The *PRINT-PRETTY* flag */

vars
	ppbroken    =   false,
	ppcol       =   0,
	ppcounting  =   false,
	pptab       =   8,
	procedure   (ppcucharout),
	;


define lconstant PP_< (ppcol, poplinemax) with_nargs 3;
	dlocal
		pc_lablist,
		ppcol,
		poplinemax,
		ppcounting      = PP_<,
		ppcucharout     = erase,
		standard_output = Null_io,
		;
	_lisppr();
	true
enddefine;


define lconstant PPcharout(char);
	lvars char;
	if ppcol fi_>= poplinemax then
		if ppcounting then
			exitfrom(false, ppcounting)
		elseunless char == `\n` do
			ppcucharout(`\n`);
			ppcucharout(`\t`);
			pptab -> ppcol;
			true -> ppbroken
		endif
	endif;
	if char == `\n` then
		0
	elseif char == `\t` then
		(ppcol fi_div pptab) fi_* pptab fi_+ pptab
	else
		ppcol fi_+ 1
	endif -> ppcol;
	ppcucharout(char)
enddefine;


define lconstant PP(level, list);
	lvars item level list save;
	dlocal depth_printed, length_printed = 0, ppbroken;

	if atom(list) or PP_<(list, level, poplinemax) then
		_lisppr(list);
		return
	elseif print_circle /== nil and Pc_printed(list) then
		return
	endif;
	CHECK_PR_LEVEL;
	list -> save;
	cucharout(`(`);
	CHECK_PR_LEN;
	fast_destpair(list) -> list -> item;
	if atom(item) then
		false -> ppbroken;
		_lisppr(item);
		if not(ppbroken) and PP_<(item, 0, 10) then
			ppcol -> level
		endif;
	else
		PP(level fi_+ 1, item);
		true -> ppbroken
	endif;
	until atom(list) or list == save do
		fast_destpair(list) -> list -> item;
		if ppbroken then
ppbroken:   cucharout(`\n`);
			while ppcol fi_< level do cucharout(`\s`) endwhile;
		endif;
		cucharout(`\s`);
		CHECK_PR_LEN;
		if atom(item) and PP_<(item, ppcol, poplinemax fi_- 1) then
			false -> ppbroken;
			_lisppr(item)
		elseif ppcol == level fi_+ 1 then
			PP(ppcol, item);
			true -> ppbroken
		else
			length_printed fi_- 1 -> length_printed;
			goto ppbroken
		endif;
	enduntil;
	if list /== [] then
		if ppbroken then
			cucharout(`\n`);
			while ppcol fi_< level do cucharout(`\s`) endwhile
		endif;
		cucharout(`\s`);
		cucharout(`.`);
		cucharout(`\s`);
		_lisppr(list)
	endif;
	cucharout(`)`)
enddefine;


define lconstant Pp_lisppr(item);
	lvars item;
	lconstant PPstandard_output = make_stream(false, PPcharout);
	dlocal
		poplinewidth    =   false,
		poplinemax,
		pptab			=	4,
		cucharout,
		ppcucharout,
		standard_output,
		;

	if isinteger(print_right_margin) and print_right_margin fi_> 0 then
		print_right_margin
 #_IF VED_LOADED
	elseif vedinvedprocess then
		vedlinemax
 #_ENDIF
	else
		76
	endif -> poplinemax;

 #_IF VED_LOADED
	if vedinvedprocess then vedindentstep -> pptab endif;
 #_ENDIF

	if isundef(ppcucharout) then
		if cucharout == charout then
			pop_charout_col
		elseif cucharout == charerr then
			pop_charerr_col
		else
			0
		endif -> ppcol;
		cucharout -> ppcucharout;
		PPcharout -> cucharout;
		PPstandard_output -> standard_output;
	endif;

	PP(ppcol, item);
enddefine;


/* The main print procedure */

define vars lisp_pr(item);
	lvars item;
	dlocal
		depth_printed   =   0,
		_lisppr         =   lispsyspr,
		pc_lablist      =   [],
		;
	SET_CUCHAROUT;
	if print_circle /== nil then
		Pc_mark(item);
		Pc_lisppr -> _lisppr
	endif;
	if print_pretty /== nil then
		Pp_lisppr(item)
	else
		_lisppr(item)
	endif
enddefine;


endsection;


/* --- Revision History ---------------------------------------------------
--- John Williams, Jun  1 1995
		Ved identifiers now guarded with #_IF VED_LOADED.
--- John Williams, May 16 1995
		Now sets poplinemax from print_right_margin.
--- John Williams, Mar 30 1995
		Changes for CLtL 2 streams.
--- John Williams, Jul 23 1993
		Changed vedediting to vedinvedprocess.
 */
