/*  --- Copyright University of Sussex 1994.  All rights reserved. ---------
 >  File:           C.x/x/ved/src/xvedscreen.p
 >  Purpose:        Defines ved settings for XVED Screen
 >  Author:         Jonathan Meyer, 20 July 1990 (see revisions)
 >  Documentation:  SYSDOC *XVED
 */
compile_mode :pop11 +strict;

section $-xved => vedxvedscreen;

#_INCLUDE 'xved_declare.ph'
include XpwScrollText.ph;
include vedscreendefs.ph;


/************************************************************************
 * Xved Screen Output
 ************************************************************************/

/* The output buffer is a fixed length string. Characters are written into
   it, and then flushed, causing XpwText widget write methods to be called.
   The buffer is also flushed whenever it becomes full. Note that flushing
   the buffer does NOT flush the X output buffer. The scrolltext widget is
   operated in non-auto-flush mode. Output flushing is performed by a read
   on the raw input device.

*/


;;; state variables

vars
	xvedscreencharmode	= 0,
	xvedscrollregion	= 0,
	xvedscreenhighlight = false,
;

lconstant macro (
	BUF_SIZE 	= 128,
);

lconstant
	buffer = writeable init_fixed(BUF_SIZE+1, string_key),
;

lvars
	buf_pos			= 1,
	curr_window		= false,
	did_char_flush	= false,
;

XptLoadProcedures xvedscreen [^^XPW_EXLIBS]
lvars
	XpwCallMethod(win,method,...),
	XFlush(dpy),
;


;;; INTERNAL PROCEDURES
;;; Flush_out - internal flush, calls methods to write characters to screen

define lconstant Flush_out(do_xflush);
	lvars do_xflush, win = curr_window;

	if wved_is_live_window(win) then
		if buf_pos /== 1 then
			0 -> fast_subscrs(buf_pos, buffer); ;;; null terminate string
			exacc [fast] (4) raw_XpwCallMethod(win,
					if vedinserting then
						XpwMInsertAtCursor
					else
						XpwMWriteAtCursor
					endif, buffer, buf_pos fi_- 1);
			true -> did_char_flush
		endif;
		if do_xflush and did_char_flush then
			exacc raw_XFlush(xveddisplay);
			false -> did_char_flush
		endif
	else
		false ->> win -> curr_window
	endif;
	1 -> buf_pos;

	if wvedwindow /== win and wved_is_live_window(wvedwindow) then
		wvedwindow -> curr_window
	endif
enddefine;


/* ======== Ved Screen Output Hooks ============================ */

define :XVED_FOR vedscr_substring_out(string, bsub, nbytes);
	lvars c, take, string, bsub, nbytes, do_write;

	dlocal pop_asts_enabled = false;
	if curr_window /== wvedwindow then Flush_out(false) endif;

	until nbytes == 0 do
		nbytes -> take; false -> do_write;
		if buf_pos fi_+ take fi_> BUF_SIZE then
			BUF_SIZE fi_- buf_pos fi_+ 1 -> take;
			true -> do_write;
		endif;
		if xvedscreencharmode &&/=_0 VEDCMODE_GRAPHIC then
			;;; graphics mode
			fast_repeat take times
				fast_subscrs(bsub, string) -> c;
				if `_` fi_<= c and c fi_<= `~` then
					if (c fi_- `_` ->> c) == 0 then `\^?` -> c endif
				endif;
				c -> fast_subscrs(buf_pos, buffer);
				bsub fi_+ 1 -> bsub;
				buf_pos fi_+ 1 -> buf_pos;
			endrepeat
		else
			move_bytes(bsub, string, buf_pos, buffer, take);
			bsub fi_+ take -> bsub;
			buf_pos fi_+ take -> buf_pos;
		endif;
		nbytes fi_- take -> nbytes;
		if do_write then Flush_out(false) endif;
	enduntil
enddefine;

define :XVED_FOR vedscr_char_out(c);
	lvars c;
	if isstring(c) then
		chain(c, 1, datalength(c), vedscr_substring_out)
	endif;

	dlocal pop_asts_enabled = false;
	if curr_window /== wvedwindow then Flush_out(false) endif;

	if xvedscreencharmode &&/=_0 VEDCMODE_GRAPHIC
	and `_` fi_<= c and c fi_<= `~` then
		if (c fi_- `_` ->> c) == 0 then `\^?` -> c endif
	endif;
	c -> fast_subscrs(buf_pos, buffer);
	buf_pos fi_+ 1 -> buf_pos;
	if buf_pos fi_> BUF_SIZE then Flush_out(false) endif
enddefine;

define :XVED_FOR vedscr_flush_output();
	dlocal pop_asts_enabled = false;
	Flush_out(true);
enddefine;


/**************************************************************************
 * Screen Control Procedures
 **************************************************************************/


/* SCROLL OPERATIONS */

define lconstant scroll(col, row, ncols, nrows, cdis, rdis);
	lvars col, row, ncols, nrows, cdis, rdis;
	dlocal pop_asts_enabled = false;
	Flush_out(false);

	;;; adjust for base 0
	col fi_- 1 -> col;
	row fi_- 1 -> row;

	if rdis /== 0 then
		;;; up/down
		if rdis fi_< 0 then
			;;; up
			row fi_- rdis -> row;
			nrows fi_+ rdis -> nrows
		else
			;;; down
			nrows fi_- rdis -> nrows
		endif;
	else
		;;; left/right
		if cdis < 0 then
			;;; left
			col fi_- cdis -> col;
			ncols fi_+ cdis -> ncols
		else
			;;; right
			ncols fi_- cdis -> ncols
		endif
	endif;
	exacc [fast] (8) raw_XpwCallMethod(wvedwindow, XpwMScroll,
									col, row, ncols, nrows, cdis, rdis)
enddefine;

define lconstant setscrollregion(top, bot) with_props vedsetscrollregion;
	lvars top, bot;
	dlocal pop_asts_enabled = false;
	Flush_out(false);
	(top fi_<< 12) fi_|| bot -> xvedscrollregion
enddefine;

define lconstant scrollregion(rdis);
	lvars rdis, top, bot;
	dlocal pop_asts_enabled = false;
	Flush_out(false);
	xvedscrollregion fi_&& 16:FFF -> bot;
	xvedscrollregion fi_>> 12 -> top;
	scroll(1, top, vedscreenwidth, bot fi_- top fi_+ 1, 0, rdis)
enddefine;
;;;
define lconstant scrollup = scrollregion(% -1 %) enddefine;
define lconstant scrolldown = scrollregion(% 1 %) enddefine;

define lconstant Scrollvertrange(srow, erow, rdis, buffline);
	lvars	n, row, incr, srow, erow, rdis, buffline,
			vbl = datalength(vedbuffer);

	until rdis == 0 do
		if rdis fi_> 0 then
			;;; scroll up
			erow, 1, fi_min(xvedvscrollstep, rdis)
		else
			;;; scroll down
			srow, -1, fi_max(-xvedvscrollstep, rdis)
		endif -> (row, incr, n);
		row fi_- n  -> row;
		rdis fi_- n -> rdis;
		scroll(1, srow, vedscreenwidth, erow fi_- srow fi_+ 1, 0, -n);
		repeat
			row fi_+ incr -> row;
			vedrefreshline(true, row, if buffline fi_> vbl then nullstring
									   else fast_subscrv(buffline, vedbuffer)
									   endif, vedmarked(buffline));
			buffline fi_+ incr -> buffline;
			quitif((n fi_- incr ->> n) == 0)
		endrepeat
	enduntil
enddefine;

define lconstant Scrollregion(srow, erow, rdis, buffline) with_props vedscrollregion;
	lvars srow, erow, rdis, buffline, tmp;
	returnunless(vedediting);
	-rdis -> rdis;		;;; has opposite sign
	if rdis fi_< 0 then
		;;; scrolling down -- the ____erow arg specifies the first line that
		;;; will be deleted (and the rdis arg can take it off the screen)
		if (srow fi_- rdis fi_- 1 fi_- vedwindowlength ->> tmp) fi_> 0 then
			rdis fi_+ tmp -> rdis
		endif;
		fi_min(erow fi_- rdis fi_- 1, vedwindowlength) -> erow;
		buffline fi_- rdis fi_- 1 -> buffline
	else
		;;; scrolling up
		if (erow fi_- rdis fi_- 1 ->> tmp) fi_< 0 then
			rdis fi_+ tmp  -> rdis;
			buffline fi_- tmp -> buffline;
			2 -> srow
		endif
	endif;
	Scrollvertrange(srow, erow, rdis, buffline)
enddefine;

define lconstant Scrollvert(nrows) with_props vedscrollvert;
	lvars nrows, newoffs = vedlineoffset fi_+ nrows;
	if newoffs fi_< 0 then
		nrows fi_- newoffs -> nrows;
		0 -> newoffs
	endif;
	Scrollvertrange(2, vedwindowlength, nrows,
						if nrows fi_> 0 then
							vedlineoffset fi_+ vedwindowlength
						else
							vedlineoffset
						endif);
	newoffs -> vedlineoffset;
	if vedlineoffset fi_+ vedwindowlength fi_>= datalength(vedbuffer) then
		vedbufferextend()
	endif
enddefine;

define lconstant Scrollhorz(ncols) with_props vedscrollhorz;
	lvars ncols, n;

	define lconstant do_scrollhorz(cdis);
		lvars cdis;
		chain(vedwcolumnoffset fi_+ 1, vedwlineoffset fi_+ 1,
				vedwlinewidth, vedwindowlength fi_- 1, cdis, 0, scroll)
	enddefine;

	define lconstant refreshrange(scol, ncols, draw_mark);
		lvars	col, scol, ncols, line, vrow = 1, coffs = vedcolumnoffset,
				clim = coffs fi_+ vedwlinewidth, vbl = datalength(vedbuffer),
				used, string, draw_mark;
		unless draw_mark then 0 -> clim endunless;
		repeat
			vrow fi_+ vedlineoffset -> line;
			quitif(line fi_> vbl);
			fast_subscrv(line, vedbuffer) -> string;
			vedusedsize(string) -> used;
			coffs fi_+ scol fi_- 1 -> col;
			unless col fi_> used then
				vedwindowpoint(vrow fi_+ vedwlineoffset, scol);
				repeat
					vedscreenoutput(
								if col == clim and col /== used then
									vedscreenmoremark
								else
									fast_subscrdstring(col, string)
								endif);
					quitif((ncols fi_- 1 ->> ncols) == 0
							or (col fi_+ 1 ->> col) fi_> used)
				endrepeat;
			endunless;
			quitif((vrow fi_+ 1 ->> vrow) == vedwindowlength)
		endrepeat
	enddefine;

	until ncols == 0 do
		if ncols fi_> 0 then
			;;; scroll left
			fi_min(xvedhscrollstep, ncols) -> n;
			do_scrollhorz(-n);
			vedcolumnoffset fi_+ n -> vedcolumnoffset;
			ncols fi_- n -> ncols;
			refreshrange(vedscreenwidth fi_- n, n fi_+ 1, ncols == 0)

		else
			;;; scroll right
			fi_min(vedcolumnoffset, fi_min(xvedhscrollstep, -ncols)) -> n;
			returnif(n == 0);
			do_scrollhorz(n);
			vedcolumnoffset fi_- n -> vedcolumnoffset;
			ncols fi_+ n -> ncols;
			refreshrange(vedwcolumnoffset fi_+ 1, n, false);
			if ncols == 0 then refreshrange(vedscreenwidth, 1, true) endif;
		endif
	enduntil
enddefine;


/* INSERTING AND DELETING LINES / CHARS */

define lconstant call_simple(method);
	lvars method;
	dlocal pop_asts_enabled = false;
	Flush_out(false);
	exacc [fast] (2) raw_XpwCallMethod(wvedwindow, method);
enddefine;

define lconstant cleartail = call_simple(% XpwMClearTailAtCursor %) enddefine;
define lconstant clear = call_simple(% XpwMClearScreen %) enddefine;

define lconstant insertline = call_simple(% XpwMInsertLineAtCursor %) enddefine;
define lconstant deleteline = call_simple(% XpwMDeleteLineAtCursor %) enddefine;
define lconstant deletechar = call_simple(% XpwMDeleteCharAtCursor %) enddefine;

define lconstant sendidseq();
	dlocal pop_asts_enabled = false;
	Flush_out(false);
	;;; '\[xved' :: ved_char_in_stream -> ved_char_in_stream;
enddefine;


/* BELL */

define lconstant bell();
	dlocal pop_asts_enabled = false;
	Flush_out(false);
	if wved_is_live_window(wvedwindow) then
		;;; safe to bell
		exacc [fast] (3) raw_XpwCallMethod(wvedwindow, XpwMBell,
									xved_value("application","BellVolume"));
	else
		;;; should work on most terminals
		rawcharout('\^G');
	endif;
enddefine;


/* CURSOR CONTROL */

define lconstant charleft = call_simple(% XpwMCursorLeft %) enddefine;
define lconstant charright = call_simple(% XpwMCursorRight %) enddefine;
define lconstant chardown = call_simple(% XpwMCursorDown %) enddefine;
define lconstant charup = call_simple(% XpwMCursorUp %) enddefine;

define lconstant screenleft();
	dlocal pop_asts_enabled = false;
	Flush_out(false);
	0 -> XptVal[fast] wvedwindow(XtN cursorColumn);
enddefine;


/* OUTPUT MODE CONTROL */

define lconstant upd_vedinserting();
	dlocal pop_asts_enabled = false;
	Flush_out(false);
	() -> vedinserting
enddefine;
;;;
define lconstant insertmode = upd_vedinserting(%true%) enddefine;
define lconstant overmode = upd_vedinserting(%false%) enddefine;

define lconstant screencharmode with_props vedscreencharmode;
	xvedscreencharmode
enddefine;
;;;
define updaterof screencharmode(mode) with_props vedscreencharmode;
	lvars mode, diff, old = xvedscreencharmode, tmp, sc;
	if xvedscreenhighlight then
		;;; xvedscreenhighlight is a colour number
		xvedscreenhighlight fi_<< VEDCMODE_COLOUR_SHIFT -> sc;
		;;; set selection colour -- xvedscreenhighlight if not already that,
		;;; 0 otherwise.
		mode -> tmp;
		mode fi_&&~~ VEDCMODE_COLOURNUM -> mode;	;;; set colour 0
		unless tmp fi_&& VEDCMODE_COLOURNUM == sc then
			mode fi_|| sc -> mode
		endunless
	endif;
	(mode ||/& old) fi_&&~~ VEDCMODE_GRAPHIC -> diff;
	mode -> xvedscreencharmode;
	returnif(diff == 0);

	dlocal pop_asts_enabled = false;
	Flush_out(false);
	returnunless(wvedwindow);	;;; wvedwindow will be false if a new window
	;;; This assumes the VEDCMODE_ char attribute flags are the same
	;;; as the XpwScrollText ones
	exacc [fast] (3) raw_XpwCallMethod(wvedwindow, XpwMSetCharAttributes,
																	mode)
enddefine;


/* OTHER CONTROLS */

define lconstant screenxy(col, row) with_props vedscreenxy;
	lvars col, row;
	dlocal pop_asts_enabled = false;
	Flush_out(false);
	unless col.isinteger and row.isinteger then
		mishap(col, row, 2, 'INVALID SCREEN POINT');
	endunless;
	exacc [fast] (4) raw_XpwCallMethod(wvedwindow, XpwMCursorTo,
											col fi_- 1, row fi_- 1);
	col -> vedscreencolumn;
	row -> vedscreenline;
enddefine;


;;; vedscreenraw - remove annoying bell when we go into raw mode
define lconstant screenraw() with_props vedscreenraw;
	1000 ->> vedscreenline -> vedscreencolumn;
enddefine;

define lconstant screencooked() with_props vedscreencooked;
	vedscreencontrol(vvedscreenovermode);
	0 -> vedscreencharmode;
	vedscr_flush_output()
enddefine;

define lconstant screencursoron() with_props vedscreencursoron;
	lvars cursor_dchar = exacc [fast] (2):uint raw_XpwCallMethod(wvedwindow,
													XpwMGetTextCursor);
	cursor_dchar /== 0 and cursor_dchar
enddefine;
;;;
define updaterof screencursoron(cursor_dchar) with_props vedscreencursoron;
	lvars cursor_dchar;
	if cursor_dchar then
		if buf_pos /== 1 then
			procedure;
				dlocal pop_asts_enabled = false;
				Flush_out(false)
			endprocedure()
		endif;
		unless isinteger(cursor_dchar) and cursor_dchar fi_> 0 then
			`O` -> cursor_dchar
		endunless
	else
		0 -> cursor_dchar
	endif;
	exacc [fast] (3) raw_XpwCallMethod(wvedwindow, XpwMSetTextCursor,
														cursor_dchar);
enddefine;


define vedxvedscreen();
	;;; general control variable assignments

	"xved" 				-> vedterminalname;
	false				-> vednokeypad;
	false				-> vedscreenwrap;

	;;; procedure values

	screenraw			-> vedscreenraw;
	screencooked		-> vedscreencooked;
	screenxy			-> vedscreenxy;
	screencursoron		-> nonactive vedscreencursoron;
	screencharmode		-> nonactive vedscreencharmode;
	setscrollregion		-> vedsetscrollregion;
	Scrollvert			-> vedscrollvert;
	Scrollhorz			-> vedscrollhorz;
	Scrollregion		-> vedscrollregion;
	;;; translation for VED standard graphic codes 16:81 - 16:9F
	vedvt100screengraphtrans -> vedscreengraphtrans;

	;;; vedset cannot cope with integers in variables,
	;;; so do the setting ourselves

	sendidseq			-> vvedscreensendidseq;
	cleartail	 		-> vvedscreencleartail;
	scrollup      		-> vvedscreenscrollup;
	scrolldown    		-> vvedscreenscrolldown;
	insertline			-> vvedscreeninsertline;
	deleteline			-> vvedscreendeleteline;
	bell			 	-> vvedscreenbell;
	charleft		 	-> vvedscreencharleft;
	charright	 		-> vvedscreencharright;
	chardown		 	-> vvedscreenchardown;
	charup			 	-> vvedscreencharup;
	clear			 	-> vvedscreenclear;
	screenleft		 	-> vvedscreenscreenleft;
	insertmode		 	-> vvedscreeninsertmode;
	overmode			-> vvedscreenovermode;
	deletechar		 	-> vvedscreendeletechar;

	;;; assign identfn to these so that tests for them = nullstring
	;;; will return false (but they aren't expected to be used,
	;;; because vedscreencharmode should be used instead).
	identfn ->> vvedscreengraphic ->> vvedscreencharnormal
			->> vvedscreencharbold ->> vvedscreencharhighlight
			-> vvedscreencharunderline;

	;;; set the following to nullstring
	vedset screen
		reset =
		setpad =
		resetpad =
		init =
	endvedset;

enddefine;
;;;
uses-by_name (vedxvedscreen);

endsection;



/* --- Revision History ---------------------------------------------------
--- Julian Clinton, Sep 12 1995
		Added check for wvedwindow false in screencharmode.
--- John Gibson, May  3 1994
		sys_s*ignals_enabled -> pop_asts_enabled
--- John Gibson, Jan  9 1994
		Changed screencursoron to get/set XtN cursorChar.
		Also fixed bug in Flush_out (wasn't clearing output buffer after
		curr_window becomes defunct).
--- John Gibson, Dec 19 1993
		Made vedscr_flush_output do XFlush
--- John Gibson, Sep  7 1992
		Changed to use XptVal
--- John Gibson, Jun  5 1992
		Added XVed version of -vedscrollregion-
--- John Gibson, Feb 13 1992
		Added use of xvedselncolour in -screencharmode-
--- John Gibson, Jan 16 1992
		Made the updater of -screencharmode- use XpwMSetCharAttributes
--- John Gibson, Dec 21 1991
		Added -vedscreencharmode- code. Corrected handling of chars in
		graphics mode.
--- John Gibson, Nov 16 1991
		Combined with xvedrawout.p (which is now redundant) and completely
		rewritten. All vved- variables are now procedures and no
		'control chars' are necessary (no longer any pretence of there being
		a device).
		Made relevant procedures immune to interrupts to ensure output for
		different windows can't get mixed up.
--- John Gibson, Oct 28 1991
		Added Scrollvert for vedscrollvert and Scrollhorz for vedscrollhorz
--- John Gibson, Sep 20 1991
		Made procedures that get reset by set_term_defaults (vdinitseq.p)
		be set by vedxvedscreen instead of being :XVED_FOR
--- John Gibson, Jul  9 1991
		Replaced vvedscreencursoron/off with active variable
		vedscreencursoron.
		Made other procedures lconstants
 */
